# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../pts/api/docs/render.pct.py.

# %% auto 0
__all__ = ['preview_docs', 'render_docs', 'foo', 'bar', 'baz', 'baz2']

# %% ../../../pts/api/docs/render.pct.py 3
from pathlib import Path
from typing import Union
from tempfile import TemporaryDirectory
import yaml
import shutil
import subprocess
import nbformat

from ..const import nblite_assets_path, format_to_file_exts
from ..utils import get_project_root_and_config, read_config, _root_path_and_config_helper
from ..config import CodeLocation, NBLiteConfig
from ..export import convert_nb, get_cell_with_directives
from .cell_docs import render_cell_doc

# %% ../../../pts/api/docs/render.pct.py 6
def process_and_remove_nbdev_directives(nb_path: Path) -> dict:
    """Processes and removes nbdev directives from a notebook."""
    nb = nbformat.read(nb_path, as_version=4)

    nb['cells'] = [get_cell_with_directives(c) for c in nb['cells']]
    proc_cells = []
    for cell in nb['cells']:
        directive_keys = [d['directive'] for d in cell['directives']]
        directive_lines = [d['cell_line'] for d in cell['directives']]
        
        if 'export' in directive_keys or 'exporti' in directive_keys: # Add rendered docstring of all function and class definitions in the cell
            doc_cell = nbformat.notebooknode.NotebookNode({
                'cell_type': 'markdown',
                'metadata': {},
                'source': render_cell_doc(cell['source']),
            })
            proc_cells.append(doc_cell)
        if any([d in directive_keys for d in ['hide', 'export', 'exporti']]): continue
        
        lines_to_remove = [i for i,dk in zip(directive_lines, directive_keys) if not dk.endswith(':')] # All quarto directive keys end with ':'
        cell['source'] = '\n'.join([l for i,l in enumerate(cell['source'].split('\n')) if i not in lines_to_remove])
        del cell['source_without_directives']
        del cell['directives']
        proc_cells.append(cell)
        
    nb['cells'] = proc_cells
    _, nb = nbformat.validator.normalize(nb)
    nbformat.write(nb, nb_path)

# %% ../../../pts/api/docs/render.pct.py 7
def convert_to_ipynb_and_copy_to_folder(dest_folder: Path, root_path: Path, cl: CodeLocation) -> Path:
    nbs_folder = root_path / cl.path
    cl_file_ext = format_to_file_exts[cl.format]
    for f in nbs_folder.glob(f"**/*"):
        if f.is_dir(): continue
        if f.parent.name.startswith('.'): continue
        
        rel_path = f.relative_to(nbs_folder)
        if any(p.startswith('_') for p in rel_path.parts): continue
        
        if f.name.endswith(cl_file_ext):
            if cl.format != "ipynb":
                file_name = f.name[:-len(cl_file_ext)-1] + ".ipynb"
                dest_path = dest_folder / rel_path.parent / file_name
                dest_path.parent.mkdir(parents=True, exist_ok=True)
                convert_nb(f, dest_path, cl.format, "ipynb")
            else:
                (dest_folder / rel_path).parent.mkdir(parents=True, exist_ok=True)
                dest_path = dest_folder / rel_path
                shutil.copy(f, dest_path)
            process_and_remove_nbdev_directives(dest_path)
        elif f.suffix in ['.md', '.qmd']:
            dest_path = dest_folder / rel_path
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(f, dest_path)

# %% ../../../pts/api/docs/render.pct.py 8
def __process_subfolder(folder_path: Path, rel_path: Path) -> dict:
    contents = {}
    contents['section'] = folder_path.name
    rel_folder_path = folder_path.relative_to(rel_path)
    sub_contents = []
    for subfolder_path in folder_path.glob('*'):
        if not subfolder_path.is_dir(): continue
        if any(p.startswith('_') for p in subfolder_path.relative_to(rel_path).parts): continue
        sub_contents.append(__process_subfolder(subfolder_path, rel_path))
    if len(sub_contents) > 0:
        contents['contents'] = [{'auto': f"{rel_folder_path}/*"}, *sub_contents]
    else:
        contents['contents'] = f"{rel_folder_path}/*"
    return contents

def build_sidebar_section(parent_path: Path) -> dict:
    """
    Recursively build sidebar YAML structure for Quarto from a directory tree.
    """
    contents = [{'auto': f"/*"}]
    for subfolder_path in parent_path.glob('*'):
        if not subfolder_path.is_dir(): continue
        if any(p.startswith('_') for p in subfolder_path.relative_to(parent_path).parts): continue
        contents.append(__process_subfolder(subfolder_path, parent_path))
    return contents

# %% ../../../pts/api/docs/render.pct.py 10
def generate_quarto_yml(docs_nbs_path: Path, src_path: Path, config: NBLiteConfig) -> dict:
    with open(nblite_assets_path / 'docs' / '_quarto.yml', 'r') as file:
        quarto_yml = yaml.safe_load(file)
    quarto_yml['website']['title'] = config.docs_title
    quarto_yml['website']['sidebar']['contents'] = build_sidebar_section(src_path)
    with open(docs_nbs_path / '_quarto.yml', 'w') as file:
        yaml.dump(quarto_yml, file)

# %% ../../../pts/api/docs/render.pct.py 12
def prepare_docs(dest_folder:Path, docs_cl:Union[str,None] = None, root_path:Union[str,None] = None, config_path:Union[str,None] = None):
    root_path, config = _root_path_and_config_helper(root_path, config_path)
    docs_cl = docs_cl or config.docs_cl
    if not docs_cl: raise ValueError("No 'docs_cl' provided and no default 'docs_cl' provided in config.")
    cl_path = root_path / config.code_locations[docs_cl].path
    if not cl_path.exists(): raise FileNotFoundError(f"Code location '{docs_cl}' not found.")
    dest_folder = Path(dest_folder)
    convert_to_ipynb_and_copy_to_folder(dest_folder, root_path, config.code_locations[docs_cl])
    generate_quarto_yml(dest_folder, cl_path, config)
    shutil.copy(nblite_assets_path / 'docs' / 'styles.css', dest_folder / 'styles.css')        

# %% ../../../pts/api/docs/render.pct.py 14
def preview_docs(docs_cl:Union[str,None] = None, root_path:Union[str,None] = None, config_path:Union[str,None] = None, verbose:bool=False):
    with TemporaryDirectory() as tmp_dir:
        prepare_docs(Path(tmp_dir), docs_cl, root_path, config_path)
        if verbose:
            subprocess.run(['quarto', 'preview'], cwd=tmp_dir)
        else:
            subprocess.run(['quarto', 'preview'], cwd=tmp_dir, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

# %% ../../../pts/api/docs/render.pct.py 16
def render_docs(output_folder:Union[str,None], docs_cl:Union[str,None] = None, root_path:Union[str,None] = None, config_path:Union[str,None] = None, verbose:bool=False):
    root_path, _ = _root_path_and_config_helper(root_path, config_path)
    doc_folder_name = Path(output_folder).name # Necessary to get the quarto printouts to be correct
    with TemporaryDirectory() as tmp_dir:
        prepare_docs(Path(tmp_dir), docs_cl, root_path, config_path)
        if verbose:
            subprocess.run(['quarto', 'render', '--output-dir', doc_folder_name], cwd=tmp_dir)
        else:
            subprocess.run(['quarto', 'render', '--output-dir', doc_folder_name], cwd=tmp_dir, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        shutil.copytree(Path(tmp_dir) / doc_folder_name, output_folder, dirs_exist_ok=True)

# %% ../../../pts/api/docs/render.pct.py 17
def foo():
    pass

def bar():
    pass

async def baz():
    pass

# %% ../../../pts/api/docs/render.pct.py 18
async def baz2():
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../pts/api/export/01_export_as_func.pct.py.

# %% auto 0
__all__ = ['export_to_lib_as_func']

# %% ../../../pts/api/export/01_export_as_func.pct.py 3
import re
import tempfile
import nbformat
from pathlib import Path
import ast

from . import convert_nb, get_nb_directives, lookup_directive, export_to_lib, get_nb_module_export_path

# %% ../../../pts/api/export/01_export_as_func.pct.py 6
def get_nb_as_py_file(nb_path: str, lib_name: str, nb_format=None):
    with tempfile.TemporaryDirectory() as tmp_dir:  # Create a temporary directory instead
        temp_lib_path = Path(tmp_dir) / lib_name
        temp_lib_path.mkdir(parents=True, exist_ok=True)
        export_to_lib(nb_path, temp_lib_path, nb_format)
        py_file_path = get_nb_module_export_path(nb_path, temp_lib_path)
        py_file_content = Path(py_file_path).read_text()
        return py_file_content
    raise Exception('Failed to get the content of the notebook as a python file')

# %% ../../../pts/api/export/01_export_as_func.pct.py 9
def get_top_exports(nb_path: str, nb_format=None):
    """
    Get the content of the notebook as a python file
    """
    directives = get_nb_directives(nb_path, nb_format)
    header_codes = [f"# %% top_export\n{d['cell']['source_without_directives']}" for d in directives if d['directive'] == 'top_export']
    return "\n\n".join(header_codes)

# %% ../../../pts/api/export/01_export_as_func.pct.py 14
def export_to_lib_as_func(nb_path: str, lib_path: str, nb_format: str = None):
    # Get the function signature from the notebook    
    directive = lookup_directive(get_nb_directives(nb_path, nb_format), 'set_func_signature')
    func_sig = directive['cell']['source_without_directives'].strip()    
    if func_sig.endswith('...'): func_sig = func_sig[:-3]
    elif func_sig.endswith('pass'): func_sig = func_sig[:-4]
    else: raise Exception('Invalid function signature')
    
    # Get the return cell from the notebook
    directive = lookup_directive(get_nb_directives(nb_path, nb_format), 'func_return')
    if directive is not None:
        return_cell = directive['cell']['source_without_directives'].strip()
        if len(return_cell.split('\n')) > 1:
            raise Exception('Return cell must contain only one line')
        return_statement = f'return {return_cell}'
    else:
        return_statement = 'return'
     
    # Get the content of the notebook as a python file
    lib_name = Path(lib_path).stem
    py_file_content = get_nb_as_py_file(nb_path, lib_name, nb_format)
    
    # Get the function header content
    header_content = get_top_exports(nb_path, nb_format)
    
    # Construct the function
    first_line = py_file_content.split('\n')[0] # The first line contains the 'AUTOGENERATED!...' comment
    func_body = "\n".join([f"    {l}" for l in py_file_content.splitlines()])
    
    py_func_file_content = f"""
{first_line}

{header_content}

{func_sig}
{func_body}
    {return_statement}
    """.strip()

    # Check the syntax
    ast.parse(py_func_file_content)

    # Export to the library
    py_file_path = get_nb_module_export_path(nb_path, lib_path)
    Path(py_file_path).parent.mkdir(parents=True, exist_ok=True)
    with open(py_file_path, 'w') as f:
        f.write(py_func_file_content)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../pts/api/docs/cell_docs.pct.py.

# %% auto 0
__all__ = ['render_cell_doc', 'show_doc']

# %% ../../../pts/api/docs/cell_docs.pct.py 3
from IPython.display import Markdown
from pathlib import Path
from docstring_parser import parse as parse_docstring
import inspect
import ast

# %% ../../../pts/api/docs/cell_docs.pct.py 7
def extract_top_level_definitions(code_str: str) -> list:
    """
    Extracts top-level function and class definitions from a given Python code string.
    """
    # Parse the code string into an Abstract Syntax Tree (AST)
    tree = ast.parse(code_str)

    # Initialize a list to store the top-level definitions
    top_level_definitions = []

    # Iterate over the top-level nodes in the AST
    for node in tree.body:
        # Check if the node is a function or class definition
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
            # Get the source code for the node
            start_line = node.lineno - 1
            end_line = node.end_lineno
            lines = code_str.splitlines()
            definition = "\n".join(lines[start_line:end_line])
            top_level_definitions.append({
                'type': 'function' if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)) else 'class',
                'code': definition,
            })

    return top_level_definitions

# %% ../../../pts/api/docs/cell_docs.pct.py 10
def extract_function_meta(code_str):
    """
    Extracts details of functions from a given Python code string.
    """
    import ast
    tree = ast.parse(code_str)
    function_details = []
    for node in tree.body:
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            func_name = node.name
            is_async = isinstance(node, ast.AsyncFunctionDef)
            args = {}
            # Regular arguments
            for arg in node.args.args:
                args[arg.arg] = ast.get_source_segment(code_str, arg.annotation) if arg.annotation else None
            # *args
            if node.args.vararg:
                vararg = node.args.vararg
                args[f"*{vararg.arg}"] = ast.get_source_segment(code_str, vararg.annotation) if vararg.annotation else None
            # **kwargs
            if node.args.kwarg:
                kwarg = node.args.kwarg
                args[f"**{kwarg.arg}"] = ast.get_source_segment(code_str, kwarg.annotation) if kwarg.annotation else None
            docstring = ast.get_docstring(node)
            # Get return type
            return_type = ast.get_source_segment(code_str, node.returns) if node.returns else None
            # Build signature string
            sig_parts = []
            for k, v in args.items():
                if v is not None:
                    sig_parts.append(f"{k}: {v}")
                else:
                    sig_parts.append(f"{k}")
            full_signature = f"{func_name}({', '.join(sig_parts)})"
            if return_type is not None:
                full_signature += f" -> {return_type}"
            function_details.append({
                'name': func_name,
                'full_signature': full_signature,
                'is_async': is_async,
                'args': args,
                'docstring': docstring,
                'return_annotation': return_type
            })
    if len(function_details) != 1: raise ValueError(f"Expected exactly one function definition in the code string. Got:\n{code_str}")
    return function_details[0]

# %% ../../../pts/api/docs/cell_docs.pct.py 14
def extract_function_meta_from_obj(func):
    """
    Extracts details of a function from a given Python function object.
    """
    import inspect

    if not inspect.isfunction(func) and not inspect.ismethod(func):
        raise TypeError("Expected a function or method object.")

    func_name = func.__name__
    sig = inspect.signature(func)
    args = {}
    sig_parts = []
    for name, param in sig.parameters.items():
        annotation = param.annotation if param.annotation is not inspect.Parameter.empty else None
        # Convert annotation to string if present
        annotation_str = (
            annotation.__name__ if isinstance(annotation, type)
            else str(annotation) if annotation is not None
            else None
        )
        # Handle *args and **kwargs
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            display_name = f"*{name}"
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            display_name = f"**{name}"
        else:
            display_name = name
        args[display_name] = annotation_str
        if annotation_str is not None:
            sig_parts.append(f"{display_name}: {annotation_str}")
        else:
            sig_parts.append(f"{display_name}")

    docstring = inspect.getdoc(func)
    is_async = inspect.iscoroutinefunction(func)

    # Get return type
    return_annotation = sig.return_annotation if sig.return_annotation is not inspect.Signature.empty else None
    return_annotation_str = (
        return_annotation.__name__ if isinstance(return_annotation, type)
        else str(return_annotation) if return_annotation is not None
        else None
    )

    full_signature = f"{func_name}({', '.join(sig_parts)})"
    if return_annotation_str is not None:
        full_signature += f" -> {return_annotation_str}"

    return {
        'name': func_name,
        'full_signature': full_signature,
        'is_async': is_async,
        'args': args,
        'docstring': docstring,
        'return_annotation': return_annotation_str
    }

# %% ../../../pts/api/docs/cell_docs.pct.py 17
def extract_class_meta(code_str):
    """
    Extracts details of a class from a given Python code string.
    """
    import ast
    tree = ast.parse(code_str)
    class_details = {}
    for node in tree.body:
        if isinstance(node, ast.ClassDef):
            class_name = node.name
            base_classes = [base.id for base in node.bases if isinstance(base, ast.Name)]
            methods = []
            for item in node.body:
                if isinstance(item, (ast.FunctionDef, ast.AsyncFunctionDef)):
                    method_details = extract_function_meta(ast.get_source_segment(code_str, item))
                    methods.append(method_details)
            class_details = {
                'name': class_name,
                'inherits_from': base_classes,
                'methods': methods
            }
    return class_details

# %% ../../../pts/api/docs/cell_docs.pct.py 20
def extract_class_meta_from_obj(cls):
    """
    Extracts details of a class from a given Python class object.
    """
    class_name = cls.__name__
    base_classes = [base.__name__ for base in cls.__bases__ if base is not object]
    methods = []
    for name, member in inspect.getmembers(cls, predicate=inspect.isfunction):
        # Only include methods defined in this class, not inherited ones
        if member.__qualname__.startswith(cls.__name__ + "."):
            methods.append(extract_function_meta_from_obj(member))
    class_details = {
        'name': class_name,
        'inherits_from': base_classes,
        'methods': methods
    }
    return class_details

# %% ../../../pts/api/docs/cell_docs.pct.py 23
def render_function_doc(func, title_level=2):
    """
    Takes function metadata and returns formatted Markdown documentation
    using `docstring-parser` to structure the docstring content.
    """
    md_lines = []

    # Header
    header = f"{'#'*title_level} {func['name']}"
    if func['is_async']:
        header += " *(async)*"
    md_lines.append(header)
    md_lines.append("")
    
    # Signature
    MAX_SIGNATURE_LENGTH = 80  # Define a constant for maximum signature length
    full_signature = func['full_signature']
    
    if len(full_signature) > MAX_SIGNATURE_LENGTH:
        # Split the signature into multiple lines
        signature_lines = [f"{func['name']}("]
        for arg, arg_type in func['args'].items():
            if arg_type:
                signature_lines.append(f"   {arg}: {arg_type},")
            else:
                signature_lines.append(f"   {arg},")
        signature_lines[-1] = signature_lines[-1].rstrip(',')  # Remove trailing comma from last argument
        if 'return_annotation' in func and func['return_annotation']:
            signature_lines.append(f") -> {func['return_annotation']}")
        else:
            signature_lines.append(")")
        if 'return_type' in func and func['return_type']:
            signature_lines[-1] += f" -> {func['return_type']}"
        md_lines.append("```python\n" + "\n".join(signature_lines) + "\n```")
    else:
        md_lines.append(f"```python\n{full_signature}\n```")
    
    md_lines.append("")

    # Parse docstring
    parsed_doc = parse_docstring(func['docstring'] or "")

    # Summary
    if parsed_doc.short_description:
        md_lines.append(parsed_doc.short_description)
        md_lines.append("")

    # Long description
    if parsed_doc.long_description:
        md_lines.append(parsed_doc.long_description)
        md_lines.append("")

    # Parameters
    if parsed_doc.params:
        md_lines.append("**Arguments:**")
        for param in parsed_doc.params:
            param_line = f"- `{param.arg_name}`"
            if param.type_name:
                param_line += f" (*{param.type_name}*)"
            if param.description:
                param_line += f": {param.description}"
            md_lines.append(param_line)
        md_lines.append("")

    # Returns
    if parsed_doc.returns:
        ret = parsed_doc.returns
        return_line = "**Returns:**"
        if ret.type_name:
            return_line += f" *{ret.type_name}*"
        if ret.description:
            return_line += f": {ret.description}"
        md_lines.append(return_line)
        md_lines.append("")

    # Final spacing
    md_lines.append("---")
    md_lines.append("")

    return "\n".join(md_lines)

# %% ../../../pts/api/docs/cell_docs.pct.py 26
def render_class_doc(cls, title_level=2):
    """
    Takes class metadata and returns formatted Markdown documentation
    for the class and its methods.
    """
    md_lines = []

    # Class header
    header = f"{'#'*title_level} {cls['name']}"
    md_lines.append(header)
    md_lines.append("")

    # Inheritance
    if cls.get('inherits_from'):
        bases = ', '.join(cls['inherits_from'])
        md_lines.append(f"*Inherits from*: `{bases}`")
        md_lines.append("")

    # Class divider
    md_lines.append("---")
    md_lines.append("")

    # Methods
    if cls.get('methods'):
        md_lines.append(f"<h{title_level+1}>Methods</h{title_level+1}>")
        md_lines.append("")
        for method in cls['methods']:
            md_lines.append(render_function_doc(method, title_level=title_level+2))

    return "\n".join(md_lines)

# %% ../../../pts/api/docs/cell_docs.pct.py 29
def render_cell_doc(cell_code, title_level=2):
    """
    Takes a cell code, extracts all top-level function and class definitions,
    and returns formatted Markdown documentation for each.
    """
    top_level_defs = extract_top_level_definitions(cell_code)
    return "\n\n".join([
        render_function_doc(extract_function_meta(func_str['code'])) if func_str['type'] == 'function' else
        render_class_doc(extract_class_meta(func_str['code']))
        for func_str in top_level_defs
    ])

# %% ../../../pts/api/docs/cell_docs.pct.py 32
def show_doc(obj, title_level=2):
    if inspect.isfunction(obj) or inspect.ismethod(obj):
        meta = extract_function_meta_from_obj(obj)
        return Markdown(render_function_doc(meta, title_level))
    elif inspect.isclass(obj):
        meta = extract_class_meta_from_obj(obj)
        return Markdown(render_class_doc(meta, title_level))
    else:
        raise ValueError("Object must be a function or class metadata dictionary.")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../pts/api/git.pct.py.

# %% auto 0
__all__ = ['has_unstaged_changes', 'get_git_root', 'list_staged_files', 'list_unstaged_and_untracked_files', 'is_file_staged',
           'get_unstaged_nb_twins']

# %% ../../pts/api/git.pct.py 3
import subprocess
from pathlib import Path

from .const import nblite_config_file_name
from .config import get_project_root_and_config, read_config
from .utils import get_code_location_nbs
from .export import get_nb_twin_paths

# %% ../../pts/api/git.pct.py 6
def has_unstaged_changes(file_path):
    """Check if the given file has unstaged changes.
    
    Args:
        file_path (str): The path to the file to check.
    
    Returns:
        bool: True if there are unstaged changes, False otherwise.
    """
    result = subprocess.run(['git', 'status', '--porcelain', file_path], capture_output=True, text=True)
    result = [l for l in result.stdout.split('\n') if l.strip()]
    assert len(result) <= 1, "Something went wrong."
    if len(result) == 0: return False
    return result[0][1] != ' '

# %% ../../pts/api/git.pct.py 9
def get_git_root():
    """
    Get the root directory of the current git repository.
    
    Returns:
        str: The path to the git root directory.
    """
    result = subprocess.run(['git', 'rev-parse', '--show-toplevel'], capture_output=True, text=True)
    return result.stdout.strip()

# %% ../../pts/api/git.pct.py 12
def list_staged_files():
    """
    List all currently staged files.
    
    Returns:
        list: A list of staged file paths.
    """
    result = subprocess.run(['git', 'diff', '--name-only', '--cached'], capture_output=True, text=True)
    return result.stdout.strip().split('\n')

# %% ../../pts/api/git.pct.py 14
def list_unstaged_and_untracked_files():
    """
    List all currently unstaged and untracked files.

    Returns:
        list: A list of unstaged and untracked file paths.
    """
    result = subprocess.run(['git', 'status', '--porcelain'], capture_output=True, text=True)
    files = []
    for line in result.stdout.strip().split('\n'):
        if line and (line[0] == ' ' or line[0] == '?'):
            files.append(line[3:])
    return files

# %% ../../pts/api/git.pct.py 17
def is_file_staged(file_path):
    """
    Check if the given file is staged.

    Args:
        file_path (str): The path to the file to check.

    Returns:
        bool: True if the file is staged, False otherwise.
    """
    result = subprocess.run(['git', 'status', '--porcelain', file_path], capture_output=True, text=True)
    result = [l for l in result.stdout.split('\n') if l.strip()]
    assert len(result) <= 1, "Something went wrong."
    if len(result) == 0: return False
    return result[0][0] == 'A' or result[0][0] == 'M'

# %% ../../pts/api/git.pct.py 20
def get_unstaged_nb_twins(root_path: str = None):
    """
    Get all notebook twins for which at least one is unstaged.
    
    Returns:
        list: A list of dictionaries, each containing 'staged' and 'unstaged' lists of twin paths.
    """
    if root_path is None:
        root_path, config = get_project_root_and_config()
    else:
        root_path = Path(root_path)
        config = read_config(root_path / nblite_config_file_name)
    git_root_path = Path(get_git_root())

    # Find all twins for which at least one is staged
    staged_twin_nbs = set()
    for cl in config.code_locations.values():
        cl_nbs = get_code_location_nbs(root_path, cl, ignore_underscores=True)
        for nb_path in cl_nbs:
            if is_file_staged(nb_path):
                staged_twin_nbs.add(get_nb_twin_paths(nb_path, root_path))

    # Check that each notebook in each twin group does not have unstaged changes
    unstaged_twins = []
    for twins in staged_twin_nbs:
        if any(has_unstaged_changes(nb_path) for nb_path in twins):
            unstaged_twins.append({
                'staged': [Path(nb_path).relative_to(git_root_path) for nb_path in twins if is_file_staged(nb_path)],
                'unstaged': [Path(nb_path).relative_to(git_root_path) for nb_path in twins if has_unstaged_changes(nb_path)],
            })
            
    return unstaged_twins
